import re
import sys
import cmd

from pengbot.utils import colorizer
from pengbot.adapters.base import BaseAdapter, UnknownCommand


class Shell(cmd.Cmd):
    def __init__(self, adapter, *args, **kwargs):
        self.adapter = adapter
        super().__init__(*args, **kwargs)
        setattr(self, 'do_%s' % self.adapter.env.bot_name, self.do_bot)

    def do_exit(self, line):
        """Exit the session"""
        with colorizer('blue'):
            print('Good bye')
        return -1

    do_EOF = do_exit
    do_bye = do_exit

    def do_directives(self, line):
        """List all directives supported by the bot"""
        for name, cmd in self.adapter.directives.items():
            with colorizer('blue'):
                print('bot %s:' % name)
                if cmd.__doc__:
                    for line in cmd.__doc__.split('\n'):
                        print('  %s' % line)
                else:
                    print()

    def says(self, message):
        with colorizer('blue'):
            print('%s> %s' % (self.adapter.env.bot_name, message))

    def do_bot(self, line):
        """Call the bot"""
        with colorizer('blue'):
            if not line:
                self.says('what?')

            try:
                res = self.adapter.receive(message=line)
            except UnknownCommand:
                self.says("I do not known what the '%s' directive is" % line)
            else:
                self.says(res)

    def cmdloop(self):
        try:
            super().cmdloop()
        except KeyboardInterrupt:
            self.do_exit(self)


class Adapter(BaseAdapter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shell = Shell(self)
        self.shell.prompt = "{env.host_user}> ".format(env=self.env)

    def receive(self, *args, **kwargs):
        with colorizer('green'):
            print('\n  Hello {env.host_user}\n'.format(env=self.env))

        self.shell.cmdloop()

    def close(self):
        return True



import logging

from .utils import colorize


class ColorizingStreamHandler(logging.StreamHandler):
    def __init__(self):
        super().__init__()

        self.level_map = {
            # LEVEL: (background, foreground, bold/normal/underscore)
            logging.DEBUG: (None, 'blue', 'normal'),
            logging.INFO: (None, 'white', 'normal'),
            logging.WARNING: (None, 'yellow', 'normal'),
            logging.ERROR: (None, 'red', 'normal'),
            logging.CRITICAL: ('red', 'white', 'bold'),
        }

    @property
    def is_tty(self):
        "Returns true if the handler's stream is a terminal."
        isatty = getattr(self.stream, 'isatty', None)
        return isatty and isatty()

    def emit(self, record):
        try:
            msg = self.format(record)
            stream = self.stream

            if record.levelno in self.level_map:
                bg, fg, style = self.level_map[record.levelno]
                with colorize(bg, fg, style, stream=self.stream):
                    stream.write(msg)
            else:
                stream.write(msg)
            stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)


def getLogger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    handler = ColorizingStreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s [%(name)s] %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



import sys
import os
import logging
import logging.config
from collections import OrderedDict
from optparse import OptionParser
from .utils import abort, imported, now

__all__ = ('main',)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(name)s:%(module)s %(process)d %(thread)d %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(asctime)s %(name)s:%(module)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        }
    },
    'loggers': {
        'pengbot': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'INFO',
        }
    }
}

logging.config.dictConfig(LOGGING)


def find_botfile(names):
    botfile_path = names.pop()

    if not (botfile_path or os.path.exists(botfile_path)):
        abort('Invalid or not existing botfile %s' % botfile_path)

    return os.path.abspath(botfile_path)


def load_features(path):
    from .command import Command
    from .listener import Listener

    with imported(path) as module:
        module_vars = vars(module)

        directives = OrderedDict()
        listeners = OrderedDict()

        docstring = module.__doc__

        for tup in module_vars.items():
            name, obj = tup

            if isinstance(obj, Command):
                directives[name] = obj

                if obj.aliases:
                    for alias in obj.aliases:
                        directives[alias] = directives[name]

            if isinstance(obj, Listener):
                listeners[name] = obj

    return docstring, directives, listeners, module


def main():
    from .api import env

    logger = logging.getLogger('pengbot')

    parser = OptionParser(usage=("%prog [options] path/to/botfile.py"), version="%prog 1.0")

    parser.add_option(
        '--loglevel', '-L',
        choices=['ERROR', 'WARNING', 'INFO', 'DEBUG'],
        dest='loglevel',
        default='INFO'
    )

    parser.add_option(
        '-n', '--name',
        dest='name',
        default='bot',
        help='name of the Bot'
    )

    parser.add_option(
        '--adapter',
        dest='adapter',
        default='shell',
        help='Bot adapter where to listen for commands, by default a interactive "shell" session'
    )

    options, arguments = parser.parse_args()
    arguments = parser.largs
    remainder_arguments = parser.rargs

    logger.setLevel(logging.getLevelName(options.loglevel))

    logger.debug('options: {}'.format(options))
    logger.debug('arguments: {}'.format(arguments))

    if len(arguments) != 1:
        parser.error("incorrect number of arguments")

    logger.debug('remainder_arguments: {}'.format(remainder_arguments))

    botfile = find_botfile(arguments)
    docstring, directives, listeners, module = load_features(botfile)

    env.module = module
    env.docstring = docstring
    env.bot_name = options.name
    env.options = options

    logger.debug('env: {}'.format(env))
    logger.debug('botfile module: {}'.format(module))
    logger.debug('directives found: {}'.format(', '.join(directives.keys())))
    logger.debug('listeners found: {}'.format(', '.join(listeners.keys())))

    adapters_path = os.path.dirname(os.path.abspath(__file__))
    adapters_path = os.path.join(adapters_path, 'adapters', '{}.py'.format(options.adapter))

    assert os.path.exists(adapters_path), 'Unknown adapter %s' % options.adapter

    with imported(adapters_path) as adapter_module:
        adapter = adapter_module.Adapter(env=env, directives=directives, listeners=listeners)

    try:
        env.bot_start_date = now()
        adapter.run()
    except SystemExit:
        raise
    except KeyboardInterrupt:
        sys.stderr.write("\nStopped.\n")
    except:
        sys.excepthook(*sys.exc_info())
        # we might leave stale threads if we don't explicitly exit()
        sys.exit(1)
    finally:
        adapter.close()

    sys.exit(0)



"""
Pengbot: For building robots with human manners
===============================================

:copyright: (c) 2016 by Mario César Señoranis Ayala.
:license: MIT, see LICENSE for more details.

"""

__title__ = 'pengbot'
__version__ = '0.1a1'
__author__ = 'Mar'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 Mario César Señoranis Ayala'

import logging
from functools import wraps

logger = logging.getLogger(__name__).addHandler(logging.NullHandler)

__all__ = ['make_robot']


def make_robot(robotClass):
    def wrapper(func):
        bot = robotClass()
        func(bot)
        return wraps(func)(bot)

    return wrapper



import re
from difflib import SequenceMatcher

__all__ = ('listener',)


class Everything:
    def __call__(self, context, message):
        return True


class FuzzyMatch:
    """
    Fuzzy String Matching.

        >>> match = FuzzyMatch('Good morning')
        >>> match({'text': 'Good morning all!'})
        ... True
        >>> match({'text': 'Good morning!'})
        ... True
        >>> match({'text': 'Good mornin'})
        ... True
    """

    def __init__(self, compare_to):
        self.compare_to = compare_to

    def __call__(self, context, message):
        match = SequenceMatcher(None, message['text'], self.compare_to)
        if match.ratio() > 0.99:
            return True


class PatternMatch:
    """
    Simple Pattern String Matching.

    #TODO: Extract matching groups if any

        >>> match = PatternMatch('#(?P<number>\d+)')
        >>> match('Issue ticket #3001')
        True
    """

    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def __call__(self, context, message):
        raise NotImplementedError()


class Command:
    """
    Command definition matching.

        >>> Command('kudos', '[<username> ...]')
        >>> Command('count', '<variable:str> <value:int> | <variable:str>')
    """

    def __init__(self, command):
        self.regexp = re.compile(r'^/%s' % command, flags=re.IGNORECASE)

    def __call__(self, context, message):
        text = message['text'].strip()

        if self.regexp.match(text):
            # TODO: Extract arguments
            return True



import asyncio
from collections import defaultdict
from functools import wraps

from pengbot import logger
from pengbot.context import Context
from pengbot.utils import BotLoggerAdapter


class UnknownCommand(Exception):
    pass


class BaseAdapter:
    handlers = []
    running = False

    def __init__(self):
        self.handlers = defaultdict(list)
        self.context = Context()
        self.logger = logger

        BotLoggerAdapter(logger, {'context': self.context})

    def __call__(self, *args, **kwargs):
        return self.cli_handler()

    @property
    def name(self):
        if self.context.get('name', None):
            return self.context['name']
        return self.__name__

    def cli_handler(self):
        # add cli options
        try:
            self.run()
        except KeyboardInterrupt:
            exit(0)

    def before_reply(self, message):
        logger.debug('before_reply: %r', message)

    def after_reply(self, message):
        logger.debug('after_reply: %r', message)

    def _done_callback(self, future):
        exc_info = future.exception()

        if exc_info:
            # handle exception
            logger.exception('Exception for %r', future, exc_info=exc_info)
        else:
            logger.debug('Done callback for %r', future)

    def receive(self, *args, **kwargs):
        raise NotImplementedError()

    def send(self, message):
        raise NotImplementedError()

    def says(self, *args, **kwargs):
        raise NotImplementedError()

    def run(self):
        loop = asyncio.get_event_loop()
        logger.info('Starting %s', self.name)
        try:
            loop.run_until_complete(self.receive())
        except KeyboardInterrupt:
            logger.info('Closing')
            loop.close()
            raise
        except Exception as err:
            logger.exception('Error', exc_info=err)
            loop.close()
            raise

    # Directives

    def hears(self, *matchs):
        def decorator(func):
            for match in matchs:
                self.logger.debug('Registering match %r for %s', match, func)

                @wraps(func)
                @asyncio.coroutine
                def callback(bot, *args, **kwargs):
                    self.logger.debug('%r: args=%r kwargs=%r', func.__name__, args, kwargs)

                    func(bot, *args, **kwargs)

                self.handlers[match].append(callback)

        return decorator

    def ask(self, question):
        def decorator(func):
            return asyncio.coroutine(func)

        return decorator

    def command(self, name):
        def decorator(func):
            return asyncio.coroutine(func)

        return decorator



from pengbot.utils import AttributeDict


class Environ:
    """Context provides attribute-style access."""

    def __init__(self, context):
        self.context = context

    def body(self):
        pass

    def handler_callback(self):
        pass

    def handler_resolver(self):
        pass


class Context(AttributeDict):
    pass



from setuptools import setup

with open('README.rst', 'r') as f:
    readme = f.read()

packages = [
    'pengbot',
    'pengbot.adapters',
]

setup(
    name='pengbot',
    description='A Bot framework to build bots with human manners',
    long_description=readme,
    version='0.1a1',
    author='Mario César Señoranis Ayala',
    author_email='mariocesar.c50@gmail.com',
    url='https://github.com/mariocesar/pengbot',
    packages=packages,
    package_data={'': ['LICENSE']},
    package_dir={'requests': 'requests'},
    entry_points={
        'console_scripts': ['pengbot = pengbot.main:main']
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Development Status :: 2 - Pre-Alpha',
    ]
)



import importlib
import json
import logging
import os
import sys
from contextlib import contextmanager
from datetime import datetime
from datetime import timedelta, tzinfo
from types import ModuleType

try:
    import pytz
except ImportError:
    pytz = None

__all__ = ('AttributeDict', 'abort', 'imported', 'utc', 'now')


class ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (list, dict, str, int, float, bool, type(None))):
            return json.JSONEncoder.default(self, obj)
        elif isinstance(obj, ModuleType):
            return obj.__file__


class AttributeDict(dict):
    """
    Dictionary subclass enabling attribute lookup/assignment of keys/values.

    For example::

        >>> m = AttributeDict({'foo': 'bar'})
        >>> m.foo
        'bar'
        >>> m.foo = 'not bar'
        >>> m['foo']
        'not bar'
    """

    def __getattr__(self, key):
        try:
            # Throws exception if not in prototype chain
            return object.__getattribute__(self, key)
        except AttributeError:
            try:
                return self[key]
            except KeyError:
                raise AttributeError

    def __setattr__(self, key, value):
        self[key] = value
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            try:
                self[key] = value
            except:
                raise AttributeError(key)
        else:
            object.__setattr__(self, key, value)

    def __delattr__(self, key):
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            try:
                del self[key]
            except KeyError:
                raise AttributeError(key)
        else:
            object.__delattr__(self, key)

    def __repr__(self):
        return 'Context(%s)' % self.__json__()

    def __str__(self):
        return self.__json__(indent=2)

    def __call__(self, **kwargs):
        print(kwargs)
        new = self.__class__(self.items())
        new.update(**kwargs)
        return new

    def __json__(self, indent=None):
        return json.dumps(self, indent=indent,
                          sort_keys=True, separators=(',', ': '),
                          cls=ObjectEncoder)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.context.pop()


def abort(message):
    import sys
    sys.stderr.write('%s\n' % message)
    exit(1)


@contextmanager
def imported(path):
    logger = logging.getLogger('pengbot')

    directory, filename = os.path.split(path)
    logger.debug('importing: %s', path)

    _added = False

    if directory not in sys.path:
        sys.path.insert(0, directory)
        _added = True

    try:
        # import_module(path/to/filename.py) -> import_module(filename)
        imported = importlib.import_module(os.path.splitext(filename)[0])
        yield imported
    finally:
        if _added:
            del sys.path[0]


color_codes = {
    'fg': {
        'black': '30',
        'red': '31',
        'green': '32',
        'yellow': '33',
        'blue': '34',
        'purple': '35',
        'magenta': '36',
        'bright gray': '37',

        'dark gray': '1;30',
        'bright red': '1;31',
        'bright green': '1;32',
        'bright yellow': '1;33',
        'bright blue': '1;34',
        'bright magenta': '1;35',
        'bright cyan': '1;36',
        'white': '1;37',
    },
    'bg': {
        'background black': '40',
        'background red': '41',
        'background green': '42',
        'background yellow': '43',
        'background blue': '44',
        'background magenta': '45',
        'background cyan': '46',
        'background white': '47',
    },
    'attr': {
        'normal': '00',
        'bold': '01',
        'underscore': '04',
        'blink': '05',
        'reverse': '06',
        'concealed': '08',
    }
}


@contextmanager
def colorize(bg=None, fg=None, attr=None, stream=None):
    """Colorize stdout"""
    stdout = stream or sys.stdout
    ansi_codes = []

    if attr:
        ansi_codes.append(color_codes['attr'][attr])
    if fg:
        ansi_codes.append(color_codes['fg'][fg])
    if bg:
        ansi_codes.append(color_codes['bg'][bg])
    stdout.write("\033[%sm" % ';'.join(ansi_codes))

    yield

    stdout.write("\033[0m")


ZERO = timedelta(0)


class UTC(tzinfo):
    """
    UTC implementation taken from Python's docs.
    Used only when pytz isn't available.
    """

    def __repr__(self):
        return "<UTC>"

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO


utc = UTC()


def now():
    return datetime.utcnow().replace(tzinfo=utc)


def isbound(obj):
    if callable(obj):
        return hasattr(obj.__call__, '__self__')
    return False


class BotLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return msg, kwargs



import asyncio
import json
import re

import websockets
from pip.utils import cached_property
from slacker import Slacker

from pengbot.adapters.base import BaseAdapter
from pengbot.matchers import PatternMatch as BasePatternMatch
from pengbot.utils import isbound

EVENTS = {
    "hello": "The client has successfully connected to the server",
    "message": "A message was sent to a channel",
    "user_typing": "A channel member is typing a message",
    "channel_marked": "Your channel read marker was updated",
    "channel_created": "A team channel was created",
    "channel_joined": "You joined a channel",
    "channel_left": "You left a channel",
    "channel_deleted": "A team channel was deleted",
    "channel_rename": "A team channel was renamed",
    "channel_archive": "A team channel was archived",
    "channel_unarchive": "A team channel was unarchived",
    "channel_history_changed": "Bulk updates were made to a channel's history",
    "dnd_updated": "Do not Disturb settings changed for the current user",
    "dnd_updated_user": "Do not Disturb settings changed for a team member",
    "im_created": "A direct message channel was created",
    "im_open": "You opened a direct message channel",
    "im_close": "You closed a direct message channel",
    "im_marked": "A direct message read marker was updated",
    "im_history_changed": "Bulk updates were made to a DM channel's history",
    "group_joined": "You joined a private group",
    "group_left": "You left a private group",
    "group_open": "You opened a group channel",
    "group_close": "You closed a group channel",
    "group_archive": "A private group was archived",
    "group_unarchive": "A private group was unarchived",
    "group_rename": "A private group was renamed",
    "group_marked": "A private group read marker was updated",
    "group_history_changed": "Bulk updates were made to a group's history",
    "file_created": "A file was created",
    "file_shared": "A file was shared",
    "file_unshared": "A file was unshared",
    "file_public": "A file was made public",
    "file_private": "A file was made private",
    "file_change": "A file was changed",
    "file_deleted": "A file was deleted",
    "file_comment_added": "A file comment was added",
    "file_comment_edited": "A file comment was edited",
    "file_comment_deleted": "A file comment was deleted",
    "pin_added": "A pin was added to a channel",
    "pin_removed": "A pin was removed from a channel",
    "presence_change": "A team member's presence changed",
    "manual_presence_change": "You manually updated your presence",
    "pref_change": "You have updated your preferences",
    "user_change": "A team member's data has changed",
    "team_join": "A new team member has joined",
    "star_added": "A team member has starred an item",
    "star_removed": "A team member removed a star",
    "reaction_added": "A team member has added an emoji reaction to an item",
    "reaction_removed": "A team member removed an emoji reaction",
    "emoji_changed": "A team custom emoji has been added or changed",
    "commands_changed": "A team slash command has been added or changed",
    "team_plan_change": "The team billing plan has changed",
    "team_pref_change": "A team preference has been updated",
    "team_rename": "The team name has changed",
    "team_domain_change": "The team domain has changed",
    "email_domain_changed": "The team email domain has changed",
    "team_profile_change": "Team profile fields have been updated",
    "team_profile_delete": "Team profile fields have been deleted",
    "team_profile_reorder": "Team profile fields have been reordered",
    "bot_added": "An integration bot was added",
    "bot_changed": "An integration bot was changed",
    "accounts_changed": "The list of accounts a user is signed into has changed",
    "team_migration_started": "The team is being migrated between servers",
    "reconnect_url": "Experimental",
    "subteam_created": "A user group has been added to the team",
    "subteam_updated": "An existing user group has been updated or its members changed",
    "subteam_self_added": "You have been added to a user group",
    "subteam_self_removed": "You have been removed from a user group",
}


class Event:
    def __init__(self, code):
        assert code in EVENTS, '%s is an unknown event' % code
        self.code = code

    def __call__(self, context, message):
        if 'type' in message:
            return message[self.code] == self.code


class Message:
    def __call__(self, context, message):
        if message.get('type', None) == 'message' and 'reply_to' not in message:
            # Dimiss messages from lost connections
            return True
        return False


class DirectMessage(Message):
    def __call__(self, context, message):
        if super().__call__(context, message):
            return message['channel'] in context['ims']


class Mention(Message):
    def __call__(self, context, message):
        if super().__call__(context, message):
            if 'text' in message:
                return '<@%s>' % context['self']['id'] in message['text']


class PatternMatch(Message, BasePatternMatch):
    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def __call__(self, context, message):
        is_message = super().__call__(context, message)
        return is_message and self.pattern.match(message['text']) is not None


class SlackRobot(BaseAdapter):
    def receive(self):
        self._message_id = 0
        resp = self.slacker.rtm.start()

        self.context.update({
            'self': resp.body['self'],
            'team': resp.body['team'],
            'users': {user['id']: user for user in resp.body['users']},
            'channels': {channel['id']: channel for channel in resp.body['channels']},
            'groups': {group['id']: group for group in resp.body['groups']},
            'ims': {i['id']: i for i in resp.body['ims']},
            'bots': resp.body['bots'],
        })
        self.logger.info('Listening to slack')
        self.logger.debug(json.dumps(self.context, indent=2))
        return self.websocket_handler(resp.body['url'])

    @cached_property
    def slacker(self):
        return Slacker(self.context.api_token)

    @asyncio.coroutine
    def websocket_handler(self, url):
        self.ws = yield from websockets.connect(url)
        self.running = True

        while True:
            raw_message = yield from self.ws.recv()

            if not raw_message:
                break

            message = json.loads(raw_message)
            self.logger.debug('Received %s', message)

            for handler_match, handler_callbacks in self.handlers.items():
                self.logger.debug('Resolving message handler %s %s', handler_match, handler_callbacks)
                self.logger.debug('Handler match is bound? %s', isbound(handler_match))

                if not isbound(handler_match):
                    handler_match = handler_match()

                if handler_match(context=self.context, message=message):
                    for callback in handler_callbacks:
                        task = asyncio.ensure_future(callback(self, message))
                        task.add_done_callback(self._done_callback)

    @asyncio.coroutine
    def send(self, message):
        content = json.dumps(message)

        yield from self.ws.send(content)

    @asyncio.coroutine
    def ping(self):
        if not self.running:
            return

        self._message_id += 1
        data = {'tries': self._message_id, 'type': 'ping'}
        content = json.dumps(data)

        yield from self.ws.send(content)

    @asyncio.coroutine
    def says(self, text, channel):
        self._message_id += 1
        data = {
            'id': self._message_id,
            'type': 'message',
            'channel': channel,
            'text': text
        }

        yield from self.send(data)
